# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""
Unit Tests for interfacing with sIPOPT and k_aug
"""

import pyomo.common.unittest as unittest

from pyomo.environ import (
    ConcreteModel,
    Objective,
    Param,
    Var,
    Block,
    Suffix,
    Constraint,
    inequality,
    NonNegativeReals,
    minimize,
    exp,
)
from pyomo.core.base.component import ComponentData
from pyomo.common.dependencies import scipy_available
from pyomo.common.collections import ComponentMap, ComponentSet
from pyomo.common.tempfiles import TempfileManager
from pyomo.core.expr.visitor import identify_variables, identify_mutable_parameters
from pyomo.contrib.sensitivity_toolbox.sens import (
    SensitivityInterface,
    _NotAnIndex,
    get_dsdp,
    get_dfds_dcds,
    line_num,
)
import pyomo.contrib.sensitivity_toolbox.examples.parameter as param_example
from pyomo.opt import SolverFactory

from pyomo.common.dependencies import (
    numpy as np,
    numpy_available,
    pandas as pd,
    pandas_available,
)
from pyomo.common.dependencies import scipy_available

opt_ipopt = SolverFactory('ipopt', solver_io='nl')
opt_kaug = SolverFactory('k_aug', solver_io='nl')
opt_dotsens = SolverFactory('dot_sens', solver_io='nl')


def make_indexed_model():
    """
    Creates the model used in the "parameter.py" example, but with indexed
    variables, parameters, and constraints.
    """
    m = ConcreteModel()

    m.x = Var([1, 2, 3], initialize={1: 0.15, 2: 0.15, 3: 0.0}, domain=NonNegativeReals)

    m.eta = Param([1, 2], initialize={1: 4.5, 2: 1.0}, mutable=True)

    m.const = Constraint(
        [1, 2],
        rule={
            1: 6 * m.x[1] + 3 * m.x[2] + 2 * m.x[3] - m.eta[1] == 0,
            2: m.eta[2] * m.x[1] + m.x[2] - m.x[3] - 1 == 0,
        },
    )

    m.cost = Objective(expr=m.x[1] ** 2 + m.x[2] ** 2 + m.x[3] ** 2)

    return m


def make_model_with_inequalities():
    """
    Creates a modified version of the model used in the "parameter.py"
    example, now with simple (one-sided) inequalities.
    """
    m = ConcreteModel()

    m.x = Var([1, 2, 3], initialize={1: 0.15, 2: 0.15, 3: 0.0}, domain=NonNegativeReals)

    m.eta = Param([1, 2], initialize={1: 4.5, 2: 1.0}, mutable=True)

    m.const = Constraint(
        [1, 2],
        rule={
            1: 6 * m.x[1] + 3 * m.x[2] + 2 * m.x[3] >= m.eta[1],
            2: m.eta[2] * m.x[1] + m.x[2] - m.x[3] - 1 <= 0,
        },
    )

    m.cost = Objective(expr=m.x[1] ** 2 + m.x[2] ** 2 + m.x[3] ** 2)

    return m


def make_model_with_ranged_inequalities():
    """
    Creates a modified version of the model used in the "parameter.py"
    example, now with simple (one-sided) inequalities.
    """
    m = ConcreteModel()

    m.x = Var([1, 2, 3], initialize={1: 0.15, 2: 0.15, 3: 0.0}, domain=NonNegativeReals)

    m.p = Param(initialize=10.0, mutable=True)

    m.eta = Param([1, 2], initialize={1: 4.5, 2: 1.0}, mutable=True)

    m.const = Constraint(
        [1, 2],
        rule={
            1: inequality(
                lower=-m.eta[1],
                body=6 * m.x[1] + 3 * m.x[2] + 2 * m.x[3],
                upper=m.p + m.eta[1],
            ),
            2: m.eta[2] * m.x[1] + m.x[2] - m.x[3] - 1 <= 0,
        },
    )

    m.cost = Objective(expr=m.x[1] ** 2 + m.x[2] ** 2 + m.x[3] ** 2)

    return m


class TestSensitivityInterface(unittest.TestCase):
    def assertIsSubset(self, s1, s2):
        for item in s1:
            self.assertIn(item, s2)

    def test_get_names(self):
        block_name = SensitivityInterface.get_default_block_name()
        self.assertEqual(block_name, "_SENSITIVITY_TOOLBOX_DATA")

        var_name = 'var'
        sens_var_name = SensitivityInterface.get_default_var_name(var_name)
        self.assertEqual(sens_var_name, var_name)

        param_name = 'param'
        sens_param_name = SensitivityInterface.get_default_param_name(param_name)
        self.assertEqual(sens_param_name, param_name)

    def test_constructor_clone(self):
        model = param_example.create_model()
        sens = SensitivityInterface(model)
        self.assertIs(sens._original_model, model)
        self.assertIsNot(sens.model_instance, model)

    def test_constructor_no_clone(self):
        model = param_example.create_model()
        sens = SensitivityInterface(model, clone_model=False)
        self.assertIs(sens._original_model, model)
        self.assertIs(sens.model_instance, model)

    def test_add_data_block(self):
        model = param_example.create_model()
        sens = SensitivityInterface(model, clone_model=False)

        block = sens._add_data_block()
        self.assertIs(sens.block.parent_block(), sens.model_instance)
        self.assertIs(sens.block.ctype, Block)
        self.assertEqual(sens.block.local_name, sens.get_default_block_name())

        with self.assertRaises(RuntimeError) as ex:
            sens._add_data_block()
        # We just tried adding the same block twice.
        self.assertIn("Cannot add component", str(ex.exception))

        # Try re-adding the same block, but this time we are prepared
        # for it to already exist.
        new_block = sens._add_data_block(existing_block=block)
        self.assertIsNot(block, new_block)

        new_block._has_replaced_expressions = True
        with self.assertRaises(RuntimeError) as ex:
            sens._add_data_block(existing_block=new_block)
        # Cannot remove and re-add sensitivity block if expressions
        # were replaced.
        self.assertIn("Re-using sensitivity interface", str(ex.exception))

    def test_process_param_list(self):
        model = make_indexed_model()
        sens = SensitivityInterface(model, clone_model=False)

        param_list = [model.x[1], model.eta]
        new_param_list = sens._process_param_list(param_list)
        self.assertIs(param_list, new_param_list)

        sens = SensitivityInterface(model, clone_model=True)
        new_param_list = sens._process_param_list(param_list)
        # The new param list contains the "same" variables in the
        # cloned model.
        self.assertIs(new_param_list[0], sens.model_instance.x[1])
        self.assertIs(new_param_list[1], sens.model_instance.eta)

    def test_add_sensitivity_data(self):
        model = make_indexed_model()
        sens = SensitivityInterface(model, clone_model=False)
        sens._add_data_block()

        param_list = [model.x, model.eta]
        with self.assertRaises(ValueError) as exc:
            sens._add_sensitivity_data(param_list)
        self.assertIn("variables must be fixed", str(exc.exception))

        sens.model_instance.x.fix()
        # NOTE: sending the same component twice probably shouldn't
        # be supported. It is convenient for testing, however.
        param_list = [model.x, model.x[1], model.eta, model.eta[1]]
        sens._add_sensitivity_data(param_list)

        block_param_list = list(sens.block.component_data_objects(Param))
        block_var_list = list(sens.block.component_data_objects(Var))
        self.assertEqual(len(block_param_list), 4)
        self.assertEqual(len(block_var_list), 3)
        self.assertEqual(len(sens.block._sens_data_list), 7)

        pred_sens_data_list = [
            (model.x[1], Param, 0, 1),
            (model.x[2], Param, 0, 2),
            (model.x[3], Param, 0, 3),
            (model.x[1], Param, 1, _NotAnIndex),
            (Var, model.eta[1], 2, 1),
            (Var, model.eta[2], 2, 2),
            (Var, model.eta[1], 3, _NotAnIndex),
        ]

        for data, pred in zip(sens.block._sens_data_list, pred_sens_data_list):
            if isinstance(pred[0], ComponentData):
                self.assertIs(data[0], pred[0])
                self.assertIs(data[1].ctype, pred[1])
                name = data[0].parent_component().local_name
                self.assertTrue(data[1].parent_component().local_name.startswith(name))
            else:
                self.assertIs(data[0].ctype, pred[0])
                self.assertIs(data[1], pred[1])
                name = data[1].parent_component().local_name
                self.assertTrue(data[0].parent_component().local_name.startswith(name))
            self.assertEqual(data[2], pred[2])
            self.assertEqual(data[3], pred[3])

    def test_expression_replacement_no_replacement(self):
        model = make_indexed_model()
        sens = SensitivityInterface(model, clone_model=False)
        sens._add_data_block()
        instance = sens.model_instance
        block = sens.block
        instance.x.fix()
        param_list = [instance.x[1], instance.x[2], instance.x[3]]
        sens._add_sensitivity_data(param_list)

        self.assertEqual(len(block.constList), 0)
        variable_sub_map = {}
        sens._replace_parameters_in_constraints(variable_sub_map)
        self.assertEqual(len(block.constList), 2)

        # Rely on order of constraints here... Fine as long as
        # component_data_objects iteration is deterministic
        pred_const_list = [instance.const[1], instance.const[2]]
        for orig, replaced in zip(pred_const_list, block.constList.values()):
            self.assertEqual(orig.expr.to_string(), replaced.expr.to_string())
            self.assertFalse(orig.active)
            self.assertTrue(replaced.active)

    def test_expression_replacement_equality(self):
        model = make_indexed_model()
        sens = SensitivityInterface(model, clone_model=False)
        sens._add_data_block()
        instance = sens.model_instance
        block = sens.block
        instance.x.fix()
        param_list = [instance.eta[1], instance.eta[2]]
        sens._add_sensitivity_data(param_list)

        orig_components = list(
            instance.component_data_objects(Constraint, active=True)
        ) + list(instance.component_data_objects(Objective, active=True))
        orig_expr = [con.expr for con in orig_components]

        # These will be modified to account for expected replacements
        expected_variables = ComponentMap(
            (con, ComponentSet(identify_variables(con.expr))) for con in orig_components
        )
        expected_parameters = ComponentMap(
            (con, ComponentSet(identify_mutable_parameters(con.expr)))
            for con in orig_components
        )

        # As constructed by the `setup_sensitivity` method:
        variable_sub_map = dict(
            (id(param), var)
            for var, param, list_idx, _ in block._sens_data_list
            if param_list[list_idx].ctype is Param
        )
        # Sanity check
        self.assertEqual(len(variable_sub_map), 2)

        # Map each param to the var that should replace it
        param_var_map = ComponentMap(
            (param, var) for var, param, _, _ in block._sens_data_list
        )

        # Remove parameters we expect to replace and add vars
        # we expect to replace with.
        for con in orig_components:
            for param in param_var_map:
                if param in expected_parameters[con]:
                    expected_variables[con].add(param_var_map[param])
                    expected_parameters[con].remove(param)

        # We check that the new components (Constraints and Objectives) contain
        # the expected parameters and variables.
        replaced = sens._replace_parameters_in_constraints(variable_sub_map)
        self.assertEqual(len(block.constList), 2)
        for con in block.constList.values():
            self.assertTrue(con.active)
            param_set = ComponentSet(identify_mutable_parameters(con.expr))
            var_set = ComponentSet(identify_variables(con.expr))
            orig_con = replaced[con]
            self.assertIsNot(orig_con, con)
            self.assertEqual(param_set, expected_parameters[orig_con])
            self.assertEqual(var_set, expected_variables[orig_con])

        self.assertIs(block.cost.ctype, Objective)
        obj = block.cost
        param_set = ComponentSet(identify_mutable_parameters(obj.expr))
        var_set = ComponentSet(identify_variables(obj.expr))
        orig_obj = replaced[obj]
        self.assertIsNot(orig_obj, obj)
        self.assertEqual(param_set, expected_parameters[orig_obj])
        self.assertEqual(var_set, expected_variables[orig_obj])

        # Original components were deactivated but otherwise not altered
        for con, expr in zip(orig_components, orig_expr):
            self.assertFalse(con.active)
            # self.assertIs(con.expr, expr)
            # ^Why does this fail?
            self.assertEqual(con.expr.to_string(), expr.to_string())

    def test_expression_replacement_inequality(self):
        model = make_model_with_inequalities()
        sens = SensitivityInterface(model, clone_model=False)
        sens._add_data_block()
        instance = sens.model_instance
        block = sens.block
        instance.x.fix()
        param_list = [instance.eta[1], instance.eta[2]]
        sens._add_sensitivity_data(param_list)

        orig_components = list(
            instance.component_data_objects(Constraint, active=True)
        ) + list(instance.component_data_objects(Objective, active=True))
        orig_expr = [con.expr for con in orig_components]

        # These will be modified to account for expected replacements
        expected_variables = ComponentMap(
            (con, ComponentSet(identify_variables(con.expr))) for con in orig_components
        )
        expected_parameters = ComponentMap(
            (con, ComponentSet(identify_mutable_parameters(con.expr)))
            for con in orig_components
        )

        # As constructed by the `setup_sensitivity` method:
        variable_sub_map = dict(
            (id(param), var)
            for var, param, list_idx, _ in block._sens_data_list
            if param_list[list_idx].ctype is Param
        )
        # Sanity check
        self.assertEqual(len(variable_sub_map), 2)

        # Map each param to the var that should replace it
        param_var_map = ComponentMap(
            (param, var) for var, param, _, _ in block._sens_data_list
        )

        # Remove parameters we expect to replace and add vars
        # we expect to replace with.
        for con in orig_components:
            for param in param_var_map:
                if param in expected_parameters[con]:
                    expected_variables[con].add(param_var_map[param])
                    expected_parameters[con].remove(param)

        # We check that the new components (Constraints and Objectives) contain
        # the expected parameters and variables.
        replaced = sens._replace_parameters_in_constraints(variable_sub_map)
        self.assertEqual(len(block.constList), 2)
        for con in block.constList.values():
            self.assertTrue(con.active)
            param_set = ComponentSet(identify_mutable_parameters(con.expr))
            var_set = ComponentSet(identify_variables(con.expr))
            orig_con = replaced[con]
            self.assertIsNot(orig_con, con)
            self.assertEqual(param_set, expected_parameters[orig_con])
            self.assertEqual(var_set, expected_variables[orig_con])

        self.assertIs(block.cost.ctype, Objective)
        obj = block.cost
        param_set = ComponentSet(identify_mutable_parameters(obj.expr))
        var_set = ComponentSet(identify_variables(obj.expr))
        orig_obj = replaced[obj]
        self.assertIsNot(orig_obj, obj)
        self.assertEqual(param_set, expected_parameters[orig_obj])
        self.assertEqual(var_set, expected_variables[orig_obj])

        # Original components were deactivated but otherwise not altered
        for con, expr in zip(orig_components, orig_expr):
            self.assertFalse(con.active)
            # self.assertIs(con.expr, expr)
            # ^Why does this fail?
            self.assertEqual(con.expr.to_string(), expr.to_string())

    def test_expression_replacement_ranged_inequality(self):
        model = make_model_with_ranged_inequalities()
        sens = SensitivityInterface(model, clone_model=False)
        sens._add_data_block()
        instance = sens.model_instance
        block = sens.block
        instance.x.fix()
        param_list = [instance.eta[1], instance.eta[2]]
        sens._add_sensitivity_data(param_list)

        orig_components = list(
            instance.component_data_objects(Constraint, active=True)
        ) + list(instance.component_data_objects(Objective, active=True))
        orig_expr = [con.expr for con in orig_components]

        # These will be modified to account for expected replacements
        expected_variables = ComponentMap(
            (con, ComponentSet(identify_variables(con.expr))) for con in orig_components
        )
        expected_parameters = ComponentMap(
            (con, ComponentSet(identify_mutable_parameters(con.expr)))
            for con in orig_components
        )

        # As constructed by the `setup_sensitivity` method:
        variable_sub_map = dict(
            (id(param), var)
            for var, param, list_idx, _ in block._sens_data_list
            if param_list[list_idx].ctype is Param
        )
        # Sanity check
        self.assertEqual(len(variable_sub_map), 2)

        # Map each param to the var that should replace it
        param_var_map = ComponentMap(
            (param, var) for var, param, _, _ in block._sens_data_list
        )

        # Remove parameters we expect to replace and add vars
        # we expect to replace with.
        for con in orig_components:
            for param in param_var_map:
                if param in expected_parameters[con]:
                    expected_variables[con].add(param_var_map[param])
                    expected_parameters[con].remove(param)

        # We check that the new components (Constraints and Objectives) contain
        # the expected parameters and variables.
        replaced = sens._replace_parameters_in_constraints(variable_sub_map)
        # With ranged inequalities, we end up with more constraints than we
        # started with:
        self.assertEqual(len(block.constList), 3)

        for con in block.constList.values():
            self.assertTrue(con.active)
            param_set = ComponentSet(identify_mutable_parameters(con.expr))
            var_set = ComponentSet(identify_variables(con.expr))
            orig_con = replaced[con]
            self.assertIsNot(orig_con, con)
            # Note that for ranged inequalities, it is not valid to check
            # that the two sets are equal as a mutable parameter could be
            # contained in only one "sub-inequality"
            self.assertIsSubset(param_set, expected_parameters[orig_con])
            self.assertEqual(var_set, expected_variables[orig_con])

        self.assertIs(block.cost.ctype, Objective)
        obj = block.cost
        param_set = ComponentSet(identify_mutable_parameters(obj.expr))
        var_set = ComponentSet(identify_variables(obj.expr))
        orig_obj = replaced[obj]
        self.assertIsNot(orig_obj, obj)
        self.assertEqual(param_set, expected_parameters[orig_obj])
        self.assertEqual(var_set, expected_variables[orig_obj])

        # Original components were deactivated but otherwise not altered
        for con, expr in zip(orig_components, orig_expr):
            self.assertFalse(con.active)
            # self.assertIs(con.expr, expr)
            # ^Why does this fail?
            self.assertEqual(con.expr.to_string(), expr.to_string())

    def test_param_const(self):
        model = make_indexed_model()
        param_list = [model.eta[1], model.eta[2]]
        sens = SensitivityInterface(model, clone_model=False)
        sens.setup_sensitivity(param_list)

        block = sens.block
        param_const = block.paramConst
        self.assertEqual(len(param_list), len(block.paramConst))

        param_var_map = ComponentMap(
            (param, var) for var, param, _, _ in block._sens_data_list
        )
        var_list = [param_var_map[param] for param in param_list]

        # Here we rely on the order of paramConst
        for param, var, con in zip(param_list, var_list, param_const.values()):
            self.assertEqual(con.body.to_string(), (var - param).to_string())

    def test_param_const_indexed(self):
        model = make_indexed_model()
        param_list = [model.eta]
        sens = SensitivityInterface(model, clone_model=False)
        sens.setup_sensitivity(param_list)

        block = sens.block
        param_const = block.paramConst

        param_var_map = ComponentMap(
            (param, var) for var, param, _, _ in block._sens_data_list
        )

        for con in param_const.values():
            var_list = list(identify_variables(con.expr))
            mut_param_list = list(identify_mutable_parameters(con.expr))
            self.assertEqual(len(var_list), 1)
            self.assertEqual(len(mut_param_list), 1)
            self.assertIs(var_list[0], param_var_map[mut_param_list[0]])
            self.assertEqual(
                con.body.to_string(), (var_list[0] - mut_param_list[0]).to_string()
            )

    def test_param_const_vars(self):
        model = make_indexed_model()
        model.x.fix()
        var_list = [model.x[2], model.x[1]]
        sens = SensitivityInterface(model, clone_model=False)
        sens.setup_sensitivity(var_list)

        block = sens.block
        param_const = block.paramConst
        self.assertEqual(len(var_list), len(block.paramConst))

        var_param_map = ComponentMap(
            (var, param) for var, param, _, _ in block._sens_data_list
        )
        param_list = [var_param_map[var] for var in var_list]

        # Here we rely on the order of paramConst
        for param, var, con in zip(param_list, var_list, param_const.values()):
            self.assertEqual(con.body.to_string(), (var - param).to_string())

    def test_suffixes_setup(self):
        model = make_indexed_model()
        param_list = [model.eta[2], model.eta[1]]
        sens = SensitivityInterface(model, clone_model=False)
        sens.setup_sensitivity(param_list)

        for i, (var, _, _, _) in enumerate(sens.block._sens_data_list):
            con = sens.block.paramConst[i + 1]
            self.assertEqual(model.sens_state_0[var], i + 1)
            self.assertEqual(model.sens_state_1[var], i + 1)
            self.assertEqual(model.sens_init_constr[con], i + 1)
            self.assertEqual(model.dcdp[con], i + 1)

        self.assertIs(type(model.sens_sol_state_1_z_L), Suffix)
        self.assertIs(type(model.sens_sol_state_1_z_U), Suffix)
        self.assertIs(type(model.ipopt_zL_out), Suffix)
        self.assertIs(type(model.ipopt_zU_out), Suffix)
        self.assertIs(type(model.ipopt_zL_in), Suffix)
        self.assertIs(type(model.ipopt_zU_in), Suffix)
        self.assertIs(type(model.dual), Suffix)
        self.assertIs(type(model.DeltaP), Suffix)

    def test_perturb_parameters_unindexed(self):
        delta = 1.0
        model = make_indexed_model()
        param_list = [model.eta[1], model.eta[2]]
        model.perturbed_eta = Param(
            [1, 2],
            mutable=True,
            initialize={i: p.value + delta for i, p in model.eta.items()},
        )
        ptb_list = [model.perturbed_eta[1], model.perturbed_eta[2]]

        sens = SensitivityInterface(model, clone_model=False)
        sens.setup_sensitivity(param_list)
        sens.perturb_parameters(ptb_list)
        instance = sens.model_instance
        block = sens.block

        param_var_map = ComponentMap(
            (param, var) for var, param, _, _ in sens.block._sens_data_list
        )
        param_con_map = ComponentMap(
            (param, block.paramConst[i + 1])
            for i, (_, param, _, _) in enumerate(sens.block._sens_data_list)
        )
        for param, ptb in zip(param_list, ptb_list):
            var = param_var_map[param]
            con = param_con_map[param]
            self.assertEqual(instance.sens_state_value_1[var], ptb.value)
            self.assertEqual(instance.DeltaP[con], -delta)

    def test_perturb_parameters_scalar(self):
        delta = 1.0
        model = make_indexed_model()
        param_list = [model.eta[1], model.eta[2]]
        model.perturbed_eta = Param(
            [1, 2],
            mutable=True,
            initialize={i: p.value + delta for i, p in model.eta.items()},
        )
        ptb_list = [model.perturbed_eta[1].value, model.perturbed_eta[2].value]

        sens = SensitivityInterface(model, clone_model=False)
        sens.setup_sensitivity(param_list)
        sens.perturb_parameters(ptb_list)
        instance = sens.model_instance
        block = sens.block

        param_var_map = ComponentMap(
            (param, var) for var, param, _, _ in sens.block._sens_data_list
        )
        param_con_map = ComponentMap(
            (param, block.paramConst[i + 1])
            for i, (_, param, _, _) in enumerate(sens.block._sens_data_list)
        )
        for param, ptb in zip(param_list, ptb_list):
            var = param_var_map[param]
            con = param_con_map[param]
            self.assertEqual(instance.sens_state_value_1[var], ptb)
            self.assertEqual(instance.DeltaP[con], -delta)

    def test_perturb_parameters_indexed(self):
        delta = 1.0
        model = make_indexed_model()
        param_list = [model.eta]
        model.perturbed_eta = Param(
            [1, 2],
            mutable=True,
            initialize={i: p.value + delta for i, p in model.eta.items()},
        )
        ptb_list = [model.perturbed_eta]

        sens = SensitivityInterface(model, clone_model=False)
        sens.setup_sensitivity(param_list)
        sens.perturb_parameters(ptb_list)
        instance = sens.model_instance
        block = sens.block

        param_var_map = ComponentMap(
            (param, var) for var, param, _, _ in sens.block._sens_data_list
        )
        param_con_map = ComponentMap(
            (param, block.paramConst[i + 1])
            for i, (_, param, _, _) in enumerate(sens.block._sens_data_list)
        )
        for param, ptb in zip(param_list, ptb_list):
            for idx in param:
                obj = param[idx]
                ptb_data = ptb[idx]
                var = param_var_map[obj]
                con = param_con_map[obj]
                self.assertEqual(instance.sens_state_value_1[var], ptb_data.value)
                self.assertEqual(instance.DeltaP[con], -delta)

    def test_perturb_indexed_parameters_with_scalar(self):
        model = make_indexed_model()
        param_list = [model.eta]
        ptb_list = [10.0]

        sens = SensitivityInterface(model, clone_model=False)
        sens.setup_sensitivity(param_list)
        sens.perturb_parameters(ptb_list)
        instance = sens.model_instance
        block = sens.block

        param_var_map = ComponentMap(
            (param, var) for var, param, _, _ in sens.block._sens_data_list
        )
        param_con_map = ComponentMap(
            (param, block.paramConst[i + 1])
            for i, (_, param, _, _) in enumerate(sens.block._sens_data_list)
        )
        for param, ptb in zip(param_list, ptb_list):
            for idx in param:
                obj = param[idx]
                var = param_var_map[obj]
                con = param_con_map[obj]
                self.assertEqual(instance.sens_state_value_1[var], ptb)
                self.assertEqual(instance.DeltaP[con], obj.value - ptb)

    @unittest.skipIf(not opt_kaug.available(False), "k_aug is not available")
    @unittest.skipIf(not opt_ipopt.available(False), "ipopt is not available")
    @unittest.skipIf(not scipy_available, "scipy is not available")
    def test_get_dsdp1(self):
        '''
        It tests the function get_dsdp with a simple nonlinear programming example.

        min f: p1*x1+ p2*(x2^2) + p1*p2
         s.t c1: x1 = p1
             c2: x2 = p2
             c3: 10 <= p1 <= 10
             c4: 5 <= p2 <= 5
        '''
        variable_name = ['p1', 'p2']

        m = ConcreteModel()
        m.x1 = Var(initialize=0)
        m.x2 = Var(initialize=0)
        m.p1 = Var(initialize=0)
        m.p2 = Var(initialize=0)
        m.obj = Objective(
            expr=m.x1 * m.p1 + m.x2 * m.x2 * m.p2 + m.p1 * m.p2, sense=minimize
        )
        m.c1 = Constraint(expr=m.x1 == m.p1)
        m.c2 = Constraint(expr=m.x2 == m.p2)
        theta = {'p1': 10.0, 'p2': 5.0}
        for v in variable_name:
            getattr(m, v).setlb(theta[v])
            getattr(m, v).setub(theta[v])
        dsdp, col = get_dsdp(m, variable_name, theta)
        ref = {'x1': [1.0, 0.0], 'x2': [0.0, 1.0], 'p1': [1.0, 0.0], 'p2': [0.0, 1.0]}
        np.testing.assert_almost_equal(
            dsdp.toarray(), np.vstack([ref[c] for c in col]).transpose()
        )

    @unittest.skipIf(not opt_kaug.available(False), "k_aug is not available")
    @unittest.skipIf(not opt_ipopt.available(False), "ipopt is not available")
    @unittest.skipIf(not scipy_available, "scipy is not available")
    def test_get_dsdp2(self):
        '''
        It tests the function get_dsdp with rooney & biegler's model.
        '''
        variable_name = ['asymptote', 'rate_constant']

        theta = {'asymptote': 19.142575284617866, 'rate_constant': 0.53109137696521}
        cov = np.array([[6.30579403, -0.4395341], [-0.4395341, 0.04193591]])
        model_uncertain = ConcreteModel()
        model_uncertain.asymptote = Var(initialize=15)
        model_uncertain.rate_constant = Var(initialize=0.5)
        model_uncertain.obj = Objective(
            expr=model_uncertain.asymptote
            * (1 - exp(-model_uncertain.rate_constant * 10)),
            sense=minimize,
        )
        theta = {'asymptote': 19.142575284617866, 'rate_constant': 0.53109137696521}
        for v in variable_name:
            getattr(model_uncertain, v).setlb(theta[v])
            getattr(model_uncertain, v).setub(theta[v])
        dsdp, col = get_dsdp(model_uncertain, variable_name, theta, {})
        np.testing.assert_almost_equal(dsdp.toarray(), [[1.0, 0.0], [0.0, 1.0]])
        assert col == ['asymptote', 'rate_constant']

    @unittest.skipIf(not opt_kaug.available(False), "k_aug is not available")
    @unittest.skipIf(not opt_dotsens.available(False), "dot_sens is not available")
    @unittest.skipIf(not scipy_available, "scipy is not available")
    def test_get_dfds_dcds(self):
        '''
        It tests the function get_sensitivity with a simple nonlinear programming example.

        min f: p1*x1+ p2*(x2^2) + p1*p2
         s.t c1: x1 = p1
             c2: x2 = p2
             c3: 10 <= p1 <= 10
             c4: 5 <= p2 <= 5
        '''
        variable_name = ['p1', 'p2']

        m = ConcreteModel()
        m.x1 = Var(initialize=0)
        m.x2 = Var(initialize=0)
        m.p1 = Var(initialize=0)
        m.p2 = Var(initialize=0)
        m.obj = Objective(
            expr=m.x1 * m.p1 + m.x2 * m.x2 * m.p2 + m.p1 * m.p2, sense=minimize
        )
        m.c1 = Constraint(expr=m.x1 == m.p1)
        m.c2 = Constraint(expr=m.x2 == m.p2)
        theta = {'p1': 10.0, 'p2': 5.0}
        for v in variable_name:
            getattr(m, v).setlb(theta[v])
            getattr(m, v).setub(theta[v])
        gradient_f, gradient_c, col, row, line_dic = get_dfds_dcds(m, variable_name)

        ref_f = {'x1': [10.0], 'x2': [50.0], 'p1': [15.0], 'p2': [35.0]}
        ref_c = {
            'x1': [1.0, 0.0],
            'x2': [0.0, 1.0],
            'p1': [-1.0, 0.0],
            'p2': [0.0, -1.0],
        }
        np.testing.assert_almost_equal(gradient_f, np.hstack([ref_f[v] for v in col]))
        np.testing.assert_almost_equal(
            gradient_c.toarray(), np.vstack([ref_c[v] for v in col]).transpose()
        )

    @unittest.skipIf(not opt_kaug.available(False), "k_aug is not available")
    @unittest.skipIf(not opt_dotsens.available(False), "dot_sens is not available")
    @unittest.skipIf(not scipy_available, "scipy is not available")
    def test_get_dfds_dcds2(self):
        '''
        It tests the function get_sensitivity with rooney & biegler's model.
        '''
        variable_name = ['asymptote', 'rate_constant']
        theta = {'asymptote': 19.142575284617866, 'rate_constant': 0.53109137696521}
        cov = np.array([[6.30579403, -0.4395341], [-0.4395341, 0.04193591]])
        model_uncertain = ConcreteModel()
        model_uncertain.asymptote = Var(initialize=15)
        model_uncertain.rate_constant = Var(initialize=0.5)
        model_uncertain.obj = Objective(
            expr=model_uncertain.asymptote
            * (1 - exp(-model_uncertain.rate_constant * 10)),
            sense=minimize,
        )
        theta = {'asymptote': 19.142575284617866, 'rate_constant': 0.53109137696521}
        for v in variable_name:
            getattr(model_uncertain, v).setlb(theta[v])
            getattr(model_uncertain, v).setub(theta[v])
        gradient_f, gradient_c, col, row, line_dic = get_dfds_dcds(
            model_uncertain, variable_name
        )
        np.testing.assert_almost_equal(gradient_f, [0.99506259, 0.945148])
        np.testing.assert_almost_equal(gradient_c, np.array([]))
        assert col == ['asymptote', 'rate_constant']
        assert row == ['obj']

    def test_line_num1(self):
        '''
        It tests the function line_num
        '''

        with TempfileManager.new_context() as TMP:
            file_name = TMP.create_tempfile('test_col.col')
            with open(file_name, "w") as file:
                file.write("var1\n")
                file.write("var3\n")
            i = line_num(file_name, 'var1')
            j = line_num(file_name, 'var3')
            self.assertEqual(i, 1)
            self.assertEqual(j, 2)

    def test_line_num2(self):
        '''
        It tests an exception error when file does not include target
        '''

        with TempfileManager.new_context() as TMP:
            file_name = TMP.create_tempfile('test_col.col')
            with open(file_name, "w") as file:
                file.write("var1\n")
                file.write("var3\n")
            with self.assertRaises(Exception) as context:
                i = line_num(file_name, 'var2')
            self.assertTrue(
                'test_col.col does not include var2' in str(context.exception)
            )


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import re

# Literals are used in parsing string names (and indicate tuples,
# indexing, and token separators)
literals = '()[],.'
# Special characters are additional characters that if they appear in
# the string force us to quote the string.  This includes the obvious
# things like single and double quote characters, but also backslash
# (indicates that the string contains escaped - possibly unicode -
# characters), and the colon (used as a token separator in the old
# ComponentUID "v1" format).
special_chars = literals + '\'":\\'

re_number = re.compile(
    r'(?:[-+]?(?:[0-9]+\.?[0-9]*|\.[0-9]+)(?:[eE][-+]?[0-9]+)?|-?inf|nan)'
)
re_special_char = re.compile(r'[' + re.escape(special_chars) + ']')


def name_repr(x, unknown_handler=str):
    if not isinstance(x, str):
        return _repr_map.get(x.__class__, unknown_handler)(x)
    else:
        x = repr(x)
        if x[1] == '|':
            return x
        unquoted = x[1:-1]
        if re_special_char.search(unquoted):
            return x
        if re_number.fullmatch(unquoted):
            return x
        return unquoted


def tuple_repr(x, unknown_handler=str):
    return (
        '('
        + ','.join(name_repr(_, unknown_handler) for _ in x)
        + (',)' if len(x) == 1 else ')')
    )


def index_repr(idx, unknown_handler=str):
    """
    Return a string representation of an index.
    """
    if idx.__class__ is tuple and len(idx) > 1:
        idx_str = ",".join(name_repr(i, unknown_handler) for i in idx)
    else:
        idx_str = name_repr(idx, unknown_handler)
    return "[" + idx_str + "]"


_repr_map = {
    slice: lambda x: '*',
    Ellipsis.__class__: lambda x: '**',
    int: repr,
    float: repr,
    str: repr,
    # Note: the function is unbound at this point; extract with __func__
    tuple: tuple_repr,
}



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import os
import os.path
import sys
import glob
import textwrap
import logging
import socket
import subprocess

import pyomo.common
from pyomo.common.collections import Bunch
from pyomo.common.tee import capture_output
import pyomo.scripting.pyomo_parser

logger = logging.getLogger('pyomo.solvers')

# --------------------------------------------------
# run
#   --list
# --------------------------------------------------


def setup_command_parser(parser):
    parser.add_argument(
        "--list",
        dest="summary",
        action='store_true',
        default=False,
        help="List the commands that are installed with Pyomo",
    )
    parser.add_argument(
        "command", nargs='*', help="The command and command-line options"
    )


def command_exec(options):
    cmddir = os.path.dirname(os.path.abspath(sys.executable)) + os.sep
    if options.summary:
        print("")
        print("The following commands are installed in the Pyomo bin directory:")
        print("----------------------------------------------------------------")
        for file in sorted(glob.glob(cmddir + '*')):
            print(" " + os.path.basename(file))
        print("")
        if len(options.command) > 0:
            print("WARNING: ignoring command specification")
        return
    if len(options.command) == 0:
        print("  ERROR: no command specified")
        return 1
    if not os.path.exists(cmddir + options.command[0]):
        print(
            "  ERROR: the command '%s' does not exist" % (cmddir + options.command[0])
        )
        return 1
    return subprocess.run(
        [cmddir] + options.command, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
    ).returncode


#
# Add a subparser for the pyomo command
#
setup_command_parser(
    pyomo.scripting.pyomo_parser.add_subparser(
        'run',
        func=command_exec,
        help='Execute a command from the Pyomo bin (or Scripts) directory.',
        description='This pyomo subcommand is used to execute commands installed with Pyomo.',
        epilog="""
This subcommand can execute any command from the bin (or Script)
directory that is created when Pyomo is installed.  Note that this
includes any commands that are installed by other Python packages
that are installed with Pyomo.  Thus, if Pyomo is installed in the
Python system directories, then this command executes any command
included with Python.
""",
    )
)

# --------------------------------------------------
# help
#   --components
#   --command
#   --transformations
#   --solvers
# --------------------------------------------------


def help_commands():
    print("")
    print("The following commands are installed with Pyomo:")
    print("-" * 75)
    registry = pyomo.common.get_pyomo_commands()
    d = max(len(key) for key in registry)
    fmt = "%%-%ds  %%s" % d
    for key in sorted(registry.keys(), key=lambda v: v.upper()):
        print(fmt % (key, registry[key]))
    print("")


def help_writers():
    import pyomo.environ
    from pyomo.opt.base import WriterFactory

    wrapper = textwrap.TextWrapper()
    wrapper.initial_indent = '      '
    wrapper.subsequent_indent = '      '
    print("")
    print("Pyomo Problem Writers")
    print("---------------------")
    for writer in sorted(WriterFactory):
        print("  " + writer)
        print(wrapper.fill(WriterFactory.doc(writer)))


def help_datamanagers(options):
    import pyomo.environ
    from pyomo.dataportal import DataManagerFactory

    wrapper = textwrap.TextWrapper()
    wrapper.initial_indent = '      '
    wrapper.subsequent_indent = '      '
    print("")
    print("Pyomo Data Managers")
    print("-------------------")
    for xform in sorted(DataManagerFactory):
        print("  " + xform)
        print(wrapper.fill(DataManagerFactory.doc(xform)))


def help_environment():
    info = Bunch()
    #
    info.python = Bunch()
    info.python.version = '%d.%d.%d' % sys.version_info[:3]
    info.python.executable = sys.executable
    info.python.platform = sys.platform
    try:
        packages = []
        import pip

        for package in pip.get_installed_distributions():
            packages.append(Bunch(name=package.project_name, version=package.version))
        info.python.packages = packages
    except:
        pass
    #
    info.environment = Bunch()
    path = os.environ.get('PATH', None)
    if not path is None:
        info.environment['shell path'] = path.split(os.pathsep)
    info.environment['python path'] = sys.path
    #
    print('#')
    print('# Information About the Python and Shell Environment')
    print('#')
    print(str(info))


def help_transformations():
    import pyomo.environ
    from pyomo.core import TransformationFactory

    wrapper = textwrap.TextWrapper()
    wrapper.initial_indent = '      '
    wrapper.subsequent_indent = '      '
    print("")
    print("Pyomo Model Transformations")
    print("---------------------------")
    for xform in sorted(TransformationFactory):
        print("  " + xform)
        _doc = TransformationFactory.doc(xform) or ""
        # Ideally, the Factory would ensure that the doc string
        # indicated deprecation, but as @deprecated() is Pyomo
        # functionality and the Factory comes directly from PyUtilib,
        # PyUtilib probably shouldn't contain Pyomo-specific processing.
        # The next best thing is to ensure that the deprecation status
        # is indicated here.
        _init_doc = TransformationFactory.get_class(xform).__init__.__doc__ or ""
        if _init_doc.strip().startswith('DEPRECATED') and 'DEPRECATE' not in _doc:
            _doc = ' '.join(('[DEPRECATED]', _doc))
        if _doc:
            print(wrapper.fill(_doc))


def help_solvers():
    import pyomo.environ

    wrapper = textwrap.TextWrapper(replace_whitespace=False)
    print("")
    print("Pyomo Solvers and Solver Managers")
    print("---------------------------------")

    print(
        wrapper.fill(
            "Pyomo uses 'solver managers' to execute 'solvers' that perform optimization and other forms of model analysis.  A solver directly executes an optimizer, typically using an executable found on the user's PATH environment.  Solver managers support a flexible mechanism for asynchronously executing solvers either locally or remotely.  The following solver managers are available in Pyomo:"
        )
    )
    print("")
    solvermgr_list = list(pyomo.opt.SolverManagerFactory)
    solvermgr_list = sorted(filter(lambda x: '_' != x[0], solvermgr_list))
    n = max(map(len, solvermgr_list))
    wrapper = textwrap.TextWrapper(subsequent_indent=' ' * (n + 9))
    for s in solvermgr_list:
        format = '    %-' + str(n) + 's     %s'
        print(wrapper.fill(format % (s, pyomo.opt.SolverManagerFactory.doc(s))))
    print("")
    wrapper = textwrap.TextWrapper(subsequent_indent='')
    print(
        wrapper.fill(
            "If no solver manager is specified, Pyomo uses the serial solver manager to execute solvers locally.  The neos solver manager is used to execute solvers on the NEOS optimization server."
        )
    )
    print("")

    print("")
    print("Serial Solver Interfaces")
    print("------------------------")
    print(wrapper.fill("The serial manager supports the following solver interfaces:"))
    print("")
    solver_list = list(pyomo.opt.SolverFactory)
    solver_list = sorted(filter(lambda x: '_' != x[0], solver_list))
    _data = []
    try:
        # Disable warnings
        logging.disable(logging.WARNING)
        # suppress ALL output
        with capture_output(capture_fd=True):
            for s in solver_list:
                # Create a solver, and see if it is available
                with pyomo.opt.SolverFactory(s) as opt:
                    ver = ''
                    if opt.available(False):
                        avail = '-'
                        if opt.license_is_valid():
                            avail = '+'
                        try:
                            ver = opt.version()
                            if isinstance(ver, str):
                                pass
                            elif ver:
                                while len(ver) > 2 and ver[-1] == 0:
                                    ver = ver[:-1]
                                ver = '.'.join(str(v) for v in ver)
                            else:
                                ver = ''
                        except (AttributeError, NameError):
                            pass
                    elif s == 'py':
                        # py is a metasolver, but since we don't specify a subsolver
                        # for this test, opt is actually an UnknownSolver, so we
                        # can't try to get the _metasolver attribute from it.
                        avail = '*'
                    elif isinstance(s, pyomo.opt.solvers.UnknownSolver):
                        # We can get here if creating a registered
                        # solver failed (i.e., an exception was raised
                        # in __init__)
                        avail = ''
                    elif getattr(opt, "_metasolver", False):
                        # Note: default to False if the attribute isn't implemented
                        avail = '*'
                    else:
                        avail = ''
                    _data.append((avail, s, ver, pyomo.opt.SolverFactory.doc(s)))
    finally:
        # Reset logging level
        logging.disable(logging.NOTSET)
    nameFieldLen = max(len(line[1]) for line in _data)
    verFieldLen = max(len(line[2]) for line in _data)
    fmt = '   %%1s%%-%ds %%-%ds %%s' % (nameFieldLen, verFieldLen)
    wrapper = textwrap.TextWrapper(
        subsequent_indent=' ' * (nameFieldLen + verFieldLen + 6)
    )
    for _line in _data:
        print(wrapper.fill(fmt % _line))

    print("")
    wrapper = textwrap.TextWrapper(subsequent_indent='')
    print(
        wrapper.fill(
            """The leading symbol (one of *, -, +) indicates the current solver availability.  A plus (+) indicates the solver is currently available to be run from Pyomo with the serial solver manager, and (if applicable) has a valid license.  A minus (-) indicates the solver executables are available but do not report having a valid license.  The solver may still be usable in an unlicensed or "demo" mode for limited problem sizes. An asterisk (*) indicates meta-solvers or generic interfaces, which are always available."""
        )
    )
    print('')
    print(
        wrapper.fill(
            'Pyomo also supports solver interfaces that are wrappers around third-party solver interfaces. These interfaces require a subsolver specification that indicates the solver being executed.  For example, the following indicates that the ipopt solver will be used:'
        )
    )
    print('')
    print('   asl:ipopt')
    print('')
    print(
        wrapper.fill(
            'The asl interface provides a generic wrapper for all solvers that use the AMPL Solver Library.'
        )
    )
    print('')
    print(
        wrapper.fill(
            'Note that subsolvers can not be enumerated automatically for these interfaces.  However, if a solver is specified that is not found, Pyomo assumes that the asl solver interface is being used.  Thus the following solver name will launch ipopt if the \'ipopt\' executable is on the user\'s path:'
        )
    )
    print('')
    print('   ipopt')
    print('')
    try:
        logging.disable(logging.WARNING)
        socket.setdefaulttimeout(10)
        import pyomo.neos.kestrel

        kestrel = pyomo.neos.kestrel.kestrelAMPL()

        solver_list = list(
            set(
                [
                    name[:-5].lower()
                    for name in kestrel.solvers()
                    if name.endswith('AMPL')
                ]
            )
        )

        print("")
        print("NEOS Solver Interfaces")
        print("----------------------")
        if len(solver_list) > 0:
            print(
                wrapper.fill(
                    "The neos solver manager supports solver interfaces that can "
                    "be executed remotely on the NEOS optimization server.  The "
                    "following solver interfaces are available with your current "
                    "system configuration:"
                )
            )
            print("")
            solver_list = sorted(solver_list)
            n = max(map(len, solver_list))
            format = '    %-' + str(n) + 's     %s'
            for name in solver_list:
                print(
                    wrapper.fill(
                        format
                        % (name, pyomo.neos.doc.get(name, 'Unexpected NEOS solver'))
                    )
                )
            print("")
        else:

            def _specific_neos_error(err):
                import socket, xmlrpc.client

                if isinstance(err, NotImplementedError):
                    return (
                        "your Python was compiled without SSL support; "
                        "HTTPS connections (required by NEOS) cannot be made."
                    )
                if isinstance(err, socket.timeout):
                    return "connection to neos-server.org timed out."
                if isinstance(err, socket.gaierror):
                    return "neos-server.org could not be resolved (DNS failure)."
                if isinstance(err, xmlrpc.client.ProtocolError):
                    return f"server responded with HTTP {err.errcode}."
                return str(err)

            reason = _specific_neos_error(getattr(kestrel, "connect_error", None))

            print(
                wrapper.fill(
                    "The neos solver manager executes solvers on the remote "
                    "NEOS optimization server, but that server is currently "
                    f"unavailable on this system.  Reason: {reason}"
                )
            )
            print("")
    except ImportError:
        pass
    finally:
        logging.disable(logging.NOTSET)
        socket.setdefaulttimeout(None)


def print_components(data):
    """
    Print information about modeling components supported by Pyomo.
    """
    from pyomo.core.base.component import ModelComponentFactory, GlobalSets

    print("")
    print("----------------------------------------------------------------")
    print("Pyomo Model Components:")
    print("----------------------------------------------------------------")
    for name in sorted(ModelComponentFactory):
        print("")
        print(" " + name)
        for line in textwrap.wrap(ModelComponentFactory.doc(name), 59):
            print("    " + line)
    print("")
    print("----------------------------------------------------------------")
    print("Pyomo Virtual Sets:")
    print("----------------------------------------------------------------")
    for name, obj in sorted(GlobalSets.items()):
        print("")
        print(" " + name)
        print("    " + obj.doc)


def help_exec(options):
    flag = False
    if options.commands:
        if options.asciidoc:
            print(
                "The '--commands' help information is not printed in an asciidoc format."
            )
        flag = True
        help_commands()
    if options.components:
        if options.asciidoc:
            print(
                "The '--components' help information is not printed in an asciidoc format."
            )
        flag = True
        print_components(None)
    if options.datamanager:
        flag = True
        help_datamanagers(options)
    if options.environment:
        flag = True
        help_environment()
    if options.transformations:
        if options.asciidoc:
            print(
                "The '--transformations' help information is not printed in an asciidoc format."
            )
        flag = True
        help_transformations()
    if options.solvers:
        if options.asciidoc:
            print(
                "The '--solvers' help information is not printed in an asciidoc format."
            )
        flag = True
        help_solvers()
    if options.writers:
        flag = True
        if options.asciidoc:
            print(
                "The '--writers' help information is not printed in an asciidoc format."
            )
        help_writers()
    if not flag:
        help_parser.print_help()


#
# Add a subparser for the pyomo command
#
def setup_help_parser(parser):
    parser.add_argument(
        "--asciidoc",
        dest="asciidoc",
        action='store_true',
        default=False,
        help="Generate output that is compatible with asciidoc's markup language",
    )
    parser.add_argument(
        "-c",
        "--commands",
        dest="commands",
        action='store_true',
        default=False,
        help="List the commands that are installed with Pyomo",
    )
    parser.add_argument(
        "--components",
        dest="components",
        action='store_true',
        default=False,
        help="List the components that are available in Pyomo's modeling environment",
    )
    parser.add_argument(
        "-d",
        "--data-managers",
        dest="datamanager",
        action='store_true',
        default=False,
        help="Print a summary of the data managers in Pyomo",
    )
    parser.add_argument(
        "-i",
        "--info",
        dest="environment",
        action='store_true',
        default=False,
        help="Summarize the environment and Python installation",
    )
    parser.add_argument(
        "-s",
        "--solvers",
        dest="solvers",
        action='store_true',
        default=False,
        help="Summarize the available solvers and solver interfaces",
    )
    parser.add_argument(
        "-t",
        "--transformations",
        dest="transformations",
        action='store_true',
        default=False,
        help="List the available model transformations",
    )
    parser.add_argument(
        "-w",
        "--writers",
        dest="writers",
        action='store_true',
        default=False,
        help="List the available problem writers",
    )
    return parser


help_parser = setup_help_parser(
    pyomo.scripting.pyomo_parser.add_subparser(
        'help',
        func=help_exec,
        help='Print help information.',
        description="This pyomo subcommand is used to print information about Pyomo's subcommands and installed Pyomo services.",
    )
)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest

from pyomo.common.log import LoggingIntercept
from pyomo.common.dependencies import numpy, numpy_available

from pyomo.core.expr.compare import assertExpressionsEqual
from pyomo.core.expr.numeric_expr import LinearExpression, MonomialTermExpression
from pyomo.core.expr import Expr_if, inequality, LinearExpression, NPV_SumExpression
import pyomo.repn.linear as linear
from pyomo.repn.linear import LinearRepn, LinearRepnVisitor
from pyomo.repn.util import InvalidNumber, OrderedVarRecorder

from pyomo.environ import (
    Any,
    ConcreteModel,
    Param,
    Var,
    Expression,
    ExternalFunction,
    cos,
    log,
)

nan = float('nan')


class VisitorConfig(dict):
    def __init__(self):
        self.subexpr = {}
        self.var_map = {}
        self.var_order = {}
        self.sorter = None
        self.var_recorder = OrderedVarRecorder(
            self.var_map, self.var_order, self.sorter
        )
        super().__init__(
            subexpression_cache=self.subexpr, var_recorder=self.var_recorder
        )

    def order_quadratic(self, quad):
        return {
            (
                (vid1, vid2)
                if self.var_order[vid1] <= self.var_order[vid2]
                else (vid2, vid1)
            ): val
            for (vid1, vid2), val in quad.items()
        }


def sum_sq(args, fixed, fgh):
    f = sum(arg**2 for arg in args)
    g = [2 * arg for arg in args]
    h = None
    return f, g, h


class TestLinear(unittest.TestCase):
    def test_finalize(self):
        m = ConcreteModel()
        m.x = Var()
        m.y = Var()
        m.z = Var()

        e = m.x + 2 * m.y - m.x - m.z

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(e)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x, id(m.y): m.y, id(m.z): m.z})
        self.assertEqual(cfg.var_order, {id(m.x): 0, id(m.y): 1, id(m.z): 2})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {id(m.y): 2, id(m.z): -1})
        self.assertEqual(repn.nonlinear, None)

        e *= 5

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(e)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x, id(m.y): m.y, id(m.z): m.z})
        self.assertEqual(cfg.var_order, {id(m.x): 0, id(m.y): 1, id(m.z): 2})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {id(m.y): 10, id(m.z): -5})
        self.assertEqual(repn.nonlinear, None)

        e = 5 * (m.y + m.z**2 + 3 * m.y**3)

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(e)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.y): m.y, id(m.z): m.z})
        self.assertEqual(cfg.var_order, {id(m.y): 0, id(m.z): 1})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {id(m.y): 5})
        assertExpressionsEqual(self, repn.nonlinear, (m.z**2 + 3 * m.y**3) * 5)

    def test_scalars(self):
        m = ConcreteModel()
        m.x = Var()
        m.p = Param(mutable=True, initialize=2)

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(3)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 3)
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.nonlinear, None)

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression((-1) ** 0.5)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        self.assertStructuredAlmostEqual(repn.constant, InvalidNumber(1j))
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.nonlinear, None)

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(m.p)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 2)
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.nonlinear, None)

        m.p.set_value(None)
        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(m.p)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, InvalidNumber(None))
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.nonlinear, None)

        m.p.set_value(nan)
        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(m.p)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        self.assertStructuredAlmostEqual(repn.constant, InvalidNumber(nan))
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.nonlinear, None)

        m.p.set_value(1j)
        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(m.p)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, InvalidNumber(1j))
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.nonlinear, None)

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(m.x)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {id(m.x): 1})
        self.assertEqual(repn.nonlinear, None)

        m.x.fix(1)
        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(m.x)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 1)
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.nonlinear, None)

        m.x.fix(None)
        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(m.x)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, InvalidNumber(None))
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.nonlinear, None)

        m.x.fix(nan)
        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(m.x)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        self.assertStructuredAlmostEqual(repn.constant, InvalidNumber(nan))
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.nonlinear, None)

        m.x.fix(1j)
        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(m.x)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, InvalidNumber(1j))
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.nonlinear, None)

    def test_npv(self):
        m = ConcreteModel()
        m.p = Param(mutable=True, initialize=4)

        nested_expr = 1 / m.p
        pow_expr = m.p ** (0.5)

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(nested_expr)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 1 / 4)
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.nonlinear, None)

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(pow_expr)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 2)
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.nonlinear, None)

        m.p = 0

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(nested_expr)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        self.assertStructuredAlmostEqual(repn.constant, InvalidNumber(nan))
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.nonlinear, None)

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(pow_expr)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.nonlinear, None)

        m.p = -1

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(nested_expr)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, -1)
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.nonlinear, None)

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(pow_expr)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        self.assertStructuredAlmostEqual(repn.constant, InvalidNumber(1j))
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.nonlinear, None)

        m.p = None

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(nested_expr)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, InvalidNumber(None))
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.nonlinear, None)

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(pow_expr)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, InvalidNumber(None))
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.nonlinear, None)

    def test_monomial(self):
        m = ConcreteModel()
        m.x = Var()
        m.p = Param(mutable=True, initialize=4)

        const_expr = 3 * m.x
        param_expr = m.p * m.x
        nested_expr = (1 / m.p) * m.x
        pow_expr = (m.p ** (0.5)) * m.x

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(const_expr)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {id(m.x): 3})
        self.assertEqual(repn.nonlinear, None)

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(param_expr)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {id(m.x): 4})
        self.assertEqual(repn.nonlinear, None)

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(nested_expr)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {id(m.x): 1 / 4})
        self.assertEqual(repn.nonlinear, None)

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(pow_expr)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {id(m.x): 2})
        self.assertEqual(repn.nonlinear, None)

        m.p = -1.0

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(param_expr)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {id(m.x): -1})
        self.assertEqual(repn.nonlinear, None)

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(nested_expr)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {id(m.x): -1})
        self.assertEqual(repn.nonlinear, None)

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(pow_expr)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertStructuredAlmostEqual(repn.linear, {id(m.x): InvalidNumber(1j)})
        self.assertEqual(repn.nonlinear, None)

        m.p = float('nan')

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(param_expr)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertStructuredAlmostEqual(repn.linear, {id(m.x): InvalidNumber(nan)})
        self.assertEqual(repn.nonlinear, None)

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(nested_expr)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertStructuredAlmostEqual(repn.linear, {id(m.x): InvalidNumber(nan)})
        self.assertEqual(repn.nonlinear, None)

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(pow_expr)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertStructuredAlmostEqual(repn.linear, {id(m.x): InvalidNumber(nan)})
        self.assertEqual(repn.nonlinear, None)

        m.p.set_value(None)

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(param_expr)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertStructuredAlmostEqual(repn.linear, {id(m.x): InvalidNumber(None)})
        self.assertEqual(repn.nonlinear, None)

        m.p.set_value(4)
        m.x.fix(10)

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(const_expr)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 30)
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.nonlinear, None)

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(param_expr)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 40)
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.nonlinear, None)

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(nested_expr)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 2.5)
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.nonlinear, None)

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(pow_expr)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 20)
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.nonlinear, None)

        m.p = float('nan')

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(param_expr)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        self.assertStructuredAlmostEqual(repn.constant, InvalidNumber(nan))
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.nonlinear, None)

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(nested_expr)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        self.assertStructuredAlmostEqual(repn.constant, InvalidNumber(nan))
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.nonlinear, None)

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(pow_expr)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        self.assertStructuredAlmostEqual(repn.constant, InvalidNumber(nan))
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.nonlinear, None)

        m.p.set_value(None)

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(param_expr)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, InvalidNumber(None))
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.nonlinear, None)

        m.p.set_value(0)
        m.x.fix(10)

        cfg = VisitorConfig()
        with LoggingIntercept() as LOG:
            repn = LinearRepnVisitor(**cfg).walk_expression(param_expr)
        self.assertEqual(LOG.getvalue(), "")

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.nonlinear, None)

        m.x.set_value(nan)

        cfg = VisitorConfig()
        with LoggingIntercept() as LOG:
            repn = LinearRepnVisitor(**cfg).walk_expression(param_expr)
        self.assertEqual(LOG.getvalue(), "")

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        self.assertStructuredAlmostEqual(repn.constant, InvalidNumber(nan))
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.nonlinear, None)

    def test_linear(self):
        m = ConcreteModel()
        m.x = Var(range(3))
        m.p = Param(mutable=True, initialize=4)

        e = LinearExpression()

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(e)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.nonlinear, None)

        e += m.x[0]

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(e)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(
            cfg.var_map, {id(m.x[0]): m.x[0], id(m.x[1]): m.x[1], id(m.x[2]): m.x[2]}
        )
        self.assertEqual(cfg.var_order, {id(m.x[0]): 0, id(m.x[1]): 1, id(m.x[2]): 2})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {id(m.x[0]): 1})
        self.assertEqual(repn.nonlinear, None)

        e += 2 * m.x[0]

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(e)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(
            cfg.var_map, {id(m.x[0]): m.x[0], id(m.x[1]): m.x[1], id(m.x[2]): m.x[2]}
        )
        self.assertEqual(cfg.var_order, {id(m.x[0]): 0, id(m.x[1]): 1, id(m.x[2]): 2})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {id(m.x[0]): 3})
        self.assertEqual(repn.nonlinear, None)

        e += m.p * m.x[1]

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(e)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(
            cfg.var_map, {id(m.x[0]): m.x[0], id(m.x[1]): m.x[1], id(m.x[2]): m.x[2]}
        )
        self.assertEqual(cfg.var_order, {id(m.x[0]): 0, id(m.x[1]): 1, id(m.x[2]): 2})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {id(m.x[0]): 3, id(m.x[1]): 4})
        self.assertEqual(repn.nonlinear, None)

        e += (m.p**0.5) * m.x[1]

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(e)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(
            cfg.var_map, {id(m.x[0]): m.x[0], id(m.x[1]): m.x[1], id(m.x[2]): m.x[2]}
        )
        self.assertEqual(cfg.var_order, {id(m.x[0]): 0, id(m.x[1]): 1, id(m.x[2]): 2})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {id(m.x[0]): 3, id(m.x[1]): 6})
        self.assertEqual(repn.nonlinear, None)

        e += 10

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(e)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(
            cfg.var_map, {id(m.x[0]): m.x[0], id(m.x[1]): m.x[1], id(m.x[2]): m.x[2]}
        )
        self.assertEqual(cfg.var_order, {id(m.x[0]): 0, id(m.x[1]): 1, id(m.x[2]): 2})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 10)
        self.assertEqual(repn.linear, {id(m.x[0]): 3, id(m.x[1]): 6})
        self.assertEqual(repn.nonlinear, None)

        e += 10 * m.p

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(e)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(
            cfg.var_map, {id(m.x[0]): m.x[0], id(m.x[1]): m.x[1], id(m.x[2]): m.x[2]}
        )
        self.assertEqual(cfg.var_order, {id(m.x[0]): 0, id(m.x[1]): 1, id(m.x[2]): 2})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 50)
        self.assertEqual(repn.linear, {id(m.x[0]): 3, id(m.x[1]): 6})
        self.assertEqual(repn.nonlinear, None)

        m.p = -1

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(e)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(
            cfg.var_map, {id(m.x[0]): m.x[0], id(m.x[1]): m.x[1], id(m.x[2]): m.x[2]}
        )
        self.assertEqual(cfg.var_order, {id(m.x[0]): 0, id(m.x[1]): 1, id(m.x[2]): 2})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertStructuredAlmostEqual(
            repn.linear, {id(m.x[0]): 3, id(m.x[1]): InvalidNumber(-1 + 1j)}
        )
        self.assertEqual(repn.nonlinear, None)

        m.p = 0
        e += (1 / m.p) * m.x[1]

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(e)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(
            cfg.var_map, {id(m.x[0]): m.x[0], id(m.x[1]): m.x[1], id(m.x[2]): m.x[2]}
        )
        self.assertEqual(cfg.var_order, {id(m.x[0]): 0, id(m.x[1]): 1, id(m.x[2]): 2})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 10)
        self.assertStructuredAlmostEqual(
            repn.linear, {id(m.x[0]): 3, id(m.x[1]): InvalidNumber(nan)}
        )
        self.assertEqual(repn.nonlinear, None)

        m.x[0].fix(10)

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(e)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x[1]): m.x[1], id(m.x[2]): m.x[2]})
        self.assertEqual(cfg.var_order, {id(m.x[0]): 0, id(m.x[1]): 1, id(m.x[2]): 2})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 40)
        self.assertStructuredAlmostEqual(repn.linear, {id(m.x[1]): InvalidNumber(nan)})
        self.assertEqual(repn.nonlinear, None)

        m.x[1].fix(10)

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(e)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        self.assertStructuredAlmostEqual(repn.constant, InvalidNumber(nan))
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.nonlinear, None)

        # Test some edge cases

        e = LinearExpression()

        e += m.x[2] + (1 / m.p)

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(e)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x[2]): m.x[2]})
        self.assertEqual(cfg.var_order, {id(m.x[0]): 0, id(m.x[1]): 1, id(m.x[2]): 2})
        self.assertEqual(repn.multiplier, 1)
        self.assertStructuredAlmostEqual(repn.constant, InvalidNumber(nan))
        self.assertEqual(repn.linear, {id(m.x[2]): 1})
        self.assertEqual(repn.nonlinear, None)

        cfg = VisitorConfig()
        cfg.var_map[id(m.x[2])] = m.x[2]
        cfg.var_map[id(m.x[0])] = m.x[0]
        cfg.var_order[id(m.x[2])] = 0
        cfg.var_order[id(m.x[0])] = 1
        repn = LinearRepnVisitor(**cfg).walk_expression(e)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x[2]): m.x[2], id(m.x[0]): m.x[0]})
        self.assertEqual(cfg.var_order, {id(m.x[2]): 0, id(m.x[0]): 1})
        self.assertEqual(repn.multiplier, 1)
        self.assertStructuredAlmostEqual(repn.constant, InvalidNumber(nan))
        self.assertEqual(repn.linear, {id(m.x[2]): 1})
        self.assertEqual(repn.nonlinear, None)

        e = LinearExpression()
        e += 0 * m.x[1]

        cfg = VisitorConfig()
        with LoggingIntercept() as LOG:
            repn = LinearRepnVisitor(**cfg).walk_expression(e)
        self.assertEqual(LOG.getvalue(), "")

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.nonlinear, None)

        m.x[1].set_value(nan)

        cfg = VisitorConfig()
        with LoggingIntercept() as LOG:
            repn = LinearRepnVisitor(**cfg).walk_expression(e)
        self.assertIn(
            "DEPRECATED: Encountered 0*InvalidNumber(nan) in expression tree.",
            LOG.getvalue(),
        )

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.nonlinear, None)

    def test_trig(self):
        m = ConcreteModel()
        m.x = Var()

        e = cos(m.x)

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(e)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {})
        assertExpressionsEqual(self, repn.nonlinear, cos(m.x))

        m.x.fix(0)

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(e)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 1)
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.nonlinear, None)

    def test_named_expr(self):
        m = ConcreteModel()
        m.x = Var(range(3))
        m.e = Expression(expr=sum((i + 2) * m.x[i] for i in range(3)))

        e = m.e * 2

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(e)
        self.assertEqual(len(cfg.subexpr), 1)
        self.assertEqual(cfg.subexpr[id(m.e)][1].multiplier, 1)
        self.assertEqual(cfg.subexpr[id(m.e)][1].constant, 0)
        self.assertEqual(
            cfg.subexpr[id(m.e)][1].linear,
            {id(m.x[0]): 2, id(m.x[1]): 3, id(m.x[2]): 4},
        )
        self.assertEqual(cfg.subexpr[id(m.e)][1].nonlinear, None)

        self.assertEqual(
            cfg.var_map, {id(m.x[0]): m.x[0], id(m.x[1]): m.x[1], id(m.x[2]): m.x[2]}
        )
        self.assertEqual(cfg.var_order, {id(m.x[0]): 0, id(m.x[1]): 1, id(m.x[2]): 2})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {id(m.x[0]): 4, id(m.x[1]): 6, id(m.x[2]): 8})
        self.assertEqual(repn.nonlinear, None)

        e = m.e * 2 + 3 * m.e

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(e)
        self.assertEqual(len(cfg.subexpr), 1)
        self.assertEqual(cfg.subexpr[id(m.e)][1].multiplier, 1)
        self.assertEqual(cfg.subexpr[id(m.e)][1].constant, 0)
        self.assertEqual(
            cfg.subexpr[id(m.e)][1].linear,
            {id(m.x[0]): 2, id(m.x[1]): 3, id(m.x[2]): 4},
        )
        self.assertEqual(cfg.subexpr[id(m.e)][1].nonlinear, None)

        self.assertEqual(
            cfg.var_map, {id(m.x[0]): m.x[0], id(m.x[1]): m.x[1], id(m.x[2]): m.x[2]}
        )
        self.assertEqual(cfg.var_order, {id(m.x[0]): 0, id(m.x[1]): 1, id(m.x[2]): 2})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {id(m.x[0]): 10, id(m.x[1]): 15, id(m.x[2]): 20})
        self.assertEqual(repn.nonlinear, None)

        m = ConcreteModel()
        m.e = Expression(expr=10)

        e = m.e * 2

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(e)
        self.assertEqual(len(cfg.subexpr), 1)
        self.assertEqual(cfg.subexpr[id(m.e)][1], 10)

        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 20)
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.nonlinear, None)

        e = m.e * 2 + 3 * m.e

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(e)
        self.assertEqual(len(cfg.subexpr), 1)
        self.assertEqual(cfg.subexpr[id(m.e)][1], 10)

        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 50)
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.nonlinear, None)

        m.e = None

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(m.e)
        self.assertEqual(
            cfg.subexpr, {id(m.e): (linear._CONSTANT, InvalidNumber(None))}
        )
        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, InvalidNumber(None))
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.nonlinear, None)

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(2 * m.e)
        self.assertEqual(
            cfg.subexpr, {id(m.e): (linear._CONSTANT, InvalidNumber(None))}
        )
        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, InvalidNumber(None))
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.nonlinear, None)

    def test_pow_expr(self):
        m = ConcreteModel()
        m.x = Var()
        m.p = Param(mutable=True, initialize=1)

        e = m.x**m.p

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(e)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {id(m.x): 1})
        self.assertEqual(repn.nonlinear, None)

        m.p = 0

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(e)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 1)
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.nonlinear, None)

        m.p = 2

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(e)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {})
        assertExpressionsEqual(self, repn.nonlinear, m.x**2)

        m.x.fix(2)

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(e)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 4)
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.nonlinear, None)

        m.p = 1 / 2
        m.x = -1

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(e)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        self.assertStructuredAlmostEqual(repn.constant, InvalidNumber(1j))
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.nonlinear, None)

        m.x.unfix()
        e = (1 + m.x) ** 2

        cfg = VisitorConfig()
        visitor = LinearRepnVisitor(**cfg)
        visitor.max_exponential_expansion = 2
        repn = visitor.walk_expression(e)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {})
        assertExpressionsEqual(self, repn.nonlinear, (m.x + 1) * (m.x + 1))

        cfg = VisitorConfig()
        visitor = LinearRepnVisitor(**cfg)
        visitor.max_exponential_expansion = 2
        visitor.expand_nonlinear_products = True
        repn = visitor.walk_expression(e)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 1)
        self.assertEqual(repn.linear, {id(m.x): 2})
        assertExpressionsEqual(self, repn.nonlinear, m.x * m.x)

    def test_product(self):
        m = ConcreteModel()
        m.x = Var()
        m.y = Var()
        m.z = Var()

        e = (2 + 3 * m.x + 4 * m.x**2) * (5 + 6 * m.x + 7 * m.x**2)

        cfg = VisitorConfig()
        visitor = LinearRepnVisitor(**cfg)
        visitor.expand_nonlinear_products = True
        repn = visitor.walk_expression(e)

        LE3 = MonomialTermExpression((3, m.x))
        LE6 = MonomialTermExpression((6, m.x))
        NL = (
            2 * (7 * m.x**2)
            + 5 * (4 * m.x**2)
            + (4 * m.x**2 + LE3) * (7 * m.x**2 + LE6)
        )

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        self.assertAlmostEqual(repn.constant, 10)
        self.assertEqual(repn.linear, {id(m.x): 27})
        assertExpressionsEqual(self, repn.nonlinear, NL)

        m.x.fix(0)
        m.y.fix(nan)
        e = m.x * m.y

        cfg = VisitorConfig()
        visitor = LinearRepnVisitor(**cfg)
        visitor.expand_nonlinear_products = True
        with LoggingIntercept() as LOG:
            repn = visitor.walk_expression(e)
        self.assertIn(
            'Encountered 0*InvalidNumber(nan) in expression tree.', LOG.getvalue()
        )

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        self.assertAlmostEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.nonlinear, None)

        e = m.x * (m.y + 2 + m.z)

        cfg = VisitorConfig()
        visitor = LinearRepnVisitor(**cfg)
        visitor.expand_nonlinear_products = True
        with LoggingIntercept() as LOG:
            repn = visitor.walk_expression(e)
        self.assertIn('Encountered 0*nan in expression tree.', LOG.getvalue())

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.z): m.z})
        self.assertEqual(cfg.var_order, {id(m.z): 0})
        self.assertEqual(repn.multiplier, 1)
        self.assertAlmostEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.nonlinear, None)

    def test_expr_if(self):
        m = ConcreteModel()
        m.x = Var()
        m.y = Var()

        e = Expr_if(m.y >= 5, m.x, m.x**2)
        f = Expr_if(m.y == 5, m.x, m.x**2)
        g = Expr_if(inequality(3, m.y, 5), m.x, m.x**2)

        m.y.fix(2)

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(e)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {})
        assertExpressionsEqual(self, repn.nonlinear, m.x**2)

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(f)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {})
        assertExpressionsEqual(self, repn.nonlinear, m.x**2)

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(g)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {})
        assertExpressionsEqual(self, repn.nonlinear, m.x**2)

        m.y.fix(5)

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(e)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {id(m.x): 1})
        self.assertEqual(repn.nonlinear, None)

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(f)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {id(m.x): 1})
        self.assertEqual(repn.nonlinear, None)

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(g)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {id(m.x): 1})
        self.assertEqual(repn.nonlinear, None)

        m.y.fix(2)
        m.x.fix(3)

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(e)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 9)
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.nonlinear, None)

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(f)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 9)
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.nonlinear, None)

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(g)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 9)
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.nonlinear, None)

        m.y.fix(5)
        m.x.fix(6)

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(e)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 6)
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.nonlinear, None)

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(f)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 6)
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.nonlinear, None)

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(g)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 6)
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.nonlinear, None)

        m.y.fix(None)
        m.x.unfix()

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(e)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {})
        assertExpressionsEqual(
            self,
            repn.nonlinear,
            Expr_if(IF=InvalidNumber(False), THEN=m.x, ELSE=m.x**2),
        )

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(f)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {})
        assertExpressionsEqual(
            self,
            repn.nonlinear,
            Expr_if(IF=InvalidNumber(False), THEN=m.x, ELSE=m.x**2),
        )

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(g)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {})
        assertExpressionsEqual(
            self,
            repn.nonlinear,
            Expr_if(IF=InvalidNumber(False), THEN=m.x, ELSE=m.x**2),
        )

        m.y.unfix()

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(e)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.y): m.y, id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.y): 0, id(m.x): 1})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {})
        assertExpressionsEqual(
            self, repn.nonlinear, Expr_if(IF=m.y >= 5, THEN=m.x, ELSE=m.x**2)
        )

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(f)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.y): m.y, id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.y): 0, id(m.x): 1})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {})
        assertExpressionsEqual(
            self, repn.nonlinear, Expr_if(IF=m.y == 5, THEN=m.x, ELSE=m.x**2)
        )

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(g)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.y): m.y, id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.y): 0, id(m.x): 1})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {})
        assertExpressionsEqual(
            self,
            repn.nonlinear,
            Expr_if(IF=inequality(3, m.y, 5), THEN=m.x, ELSE=m.x**2),
        )

        h = Expr_if(1 / m.y >= 1, m.x, m.x**2)

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(h)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x, id(m.y): m.y})
        self.assertEqual(cfg.var_order, {id(m.y): 0, id(m.x): 1})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {})
        assertExpressionsEqual(
            self, repn.nonlinear, Expr_if(IF=1 / m.y >= 1, THEN=m.x, ELSE=m.x**2)
        )

        m.y.fix(0)
        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(h)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {})
        assertExpressionsEqual(
            self,
            repn.nonlinear,
            Expr_if(IF=InvalidNumber(False), THEN=m.x, ELSE=m.x**2),
        )

    def test_division(self):
        m = ConcreteModel()
        m.x = Var()
        m.y = Var()

        e = (2 * m.x + 1) / m.y
        m.y.fix(2)

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(e)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 1 / 2)
        self.assertEqual(repn.linear, {id(m.x): 1})
        self.assertEqual(repn.nonlinear, None)

        e = m.y / (m.x + 1)

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(e)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {})
        assertExpressionsEqual(self, repn.nonlinear, 2 / (m.x + 1))

    def test_negation(self):
        m = ConcreteModel()
        m.x = Var()
        m.y = Var()

        e = -(m.x + 2)

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(e)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, -2)
        self.assertEqual(repn.linear, {id(m.x): -1})
        self.assertEqual(repn.nonlinear, None)

        m.x.fix(3)

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(e)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, -5)
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.nonlinear, None)

    def test_external(self):
        m = ConcreteModel()
        m.x = Var()
        m.y = Var()
        m.sq = ExternalFunction(fgh=sum_sq)

        e = m.sq(2 / m.x, 2 * m.y)

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(e)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {})
        self.assertIs(repn.nonlinear, e)

        m.x.fix(2)
        m.y.fix(3)

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(e)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 37)
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.nonlinear, None)

        m.x.fix(0)

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(e)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {})
        self.assertIs(repn.nonlinear, e)

    def test_errors_propagate_nan(self):
        m = ConcreteModel()
        m.p = Param(mutable=True, initialize=0, domain=Any)
        m.x = Var()
        m.y = Var()
        m.z = Var()
        m.y.fix(1)

        expr = (m.x + 1) / m.p
        cfg = VisitorConfig()
        with LoggingIntercept() as LOG:
            repn = LinearRepnVisitor(**cfg).walk_expression(expr)
        self.assertEqual(
            LOG.getvalue(),
            "Exception encountered evaluating expression 'div(1, 0)'\n"
            "\tmessage: division by zero\n"
            "\texpression: (x + 1)/p\n",
        )
        self.assertEqual(repn.multiplier, 1)
        self.assertStructuredAlmostEqual(repn.constant, InvalidNumber(nan))
        self.assertStructuredAlmostEqual(repn.linear, {id(m.x): InvalidNumber(nan)})
        self.assertEqual(repn.nonlinear, None)

        expr = m.y + m.x + m.z + ((3 * m.x) / m.p) / m.y
        cfg = VisitorConfig()
        with LoggingIntercept() as LOG:
            repn = LinearRepnVisitor(**cfg).walk_expression(expr)
        self.assertEqual(
            LOG.getvalue(),
            "Exception encountered evaluating expression 'div(3, 0)'\n"
            "\tmessage: division by zero\n"
            "\texpression: 3/p\n",
        )
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 1)
        self.assertStructuredAlmostEqual(
            repn.linear, {id(m.z): 1, id(m.x): InvalidNumber(nan)}
        )
        self.assertEqual(repn.nonlinear, None)

        m.y.fix(None)
        expr = log(m.y) + 3
        repn = LinearRepnVisitor(**cfg).walk_expression(expr)
        self.assertEqual(repn.multiplier, 1)
        self.assertStructuredAlmostEqual(repn.constant, InvalidNumber(nan))
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.nonlinear, None)

        expr = 3 * m.y
        repn = LinearRepnVisitor(**cfg).walk_expression(expr)
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, InvalidNumber(None))
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.nonlinear, None)

        m.p.value = None
        expr = 5 * (m.p * m.x + 2 * m.z)
        repn = LinearRepnVisitor(**cfg).walk_expression(expr)
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(len(repn.linear), 2)
        self.assertEqual(repn.linear[id(m.z)], 10)
        self.assertEqual(repn.linear[id(m.x)], InvalidNumber(None))
        self.assertEqual(repn.nonlinear, None)

        expr = m.y * m.x
        repn = LinearRepnVisitor(**cfg).walk_expression(expr)
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(len(repn.linear), 1)
        self.assertEqual(repn.linear[id(m.x)], InvalidNumber(None))
        self.assertEqual(repn.nonlinear, None)

        m.z = Var([1, 2, 3, 4], initialize=lambda m, i: i - 1)
        m.z[1].fix(None)
        expr = m.z[1] - ((m.z[2] * m.z[3]) * m.z[4])
        repn = LinearRepnVisitor(**cfg).walk_expression(expr)
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, InvalidNumber(None))
        self.assertEqual(repn.linear, {})
        self.assertIsNotNone(repn.nonlinear)

        m.z[3].fix(float('nan'))
        repn = LinearRepnVisitor(**cfg).walk_expression(expr)
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, InvalidNumber(None))
        self.assertEqual(repn.linear, {})
        self.assertIsNotNone(repn.nonlinear)

    def test_type_registrations(self):
        m = ConcreteModel()

        cfg = VisitorConfig()
        visitor = LinearRepnVisitor(**cfg)

        _orig_dispatcher = visitor.before_child_dispatcher
        linear._before_child_dispatcher = bcd = _orig_dispatcher.__class__()
        bcd.clear()
        try:
            # native type
            self.assertEqual(
                bcd.register_dispatcher(visitor, 5), (False, (linear._CONSTANT, 5))
            )
            self.assertEqual(len(bcd), 1)
            self.assertIs(bcd[int], bcd._before_native_numeric)
            # complex type
            self.assertEqual(
                bcd.register_dispatcher(visitor, 5j), (False, (linear._CONSTANT, 5j))
            )
            self.assertEqual(len(bcd), 2)
            self.assertIs(bcd[complex], bcd._before_complex)
            # ScalarParam
            m.p = Param(initialize=5)
            self.assertEqual(
                bcd.register_dispatcher(visitor, m.p), (False, (linear._CONSTANT, 5))
            )
            self.assertEqual(len(bcd), 3)
            self.assertIs(bcd[m.p.__class__], bcd._before_param)
            # ParamData
            m.q = Param([0], initialize=6, mutable=True)
            self.assertEqual(
                bcd.register_dispatcher(visitor, m.q[0]), (False, (linear._CONSTANT, 6))
            )
            self.assertEqual(len(bcd), 4)
            self.assertIs(bcd[m.q[0].__class__], bcd._before_param)
            # NPV_SumExpression
            self.assertEqual(
                bcd.register_dispatcher(visitor, m.p + m.q[0]),
                (False, (linear._CONSTANT, 11)),
            )
            self.assertEqual(len(bcd), 6)
            self.assertIs(bcd[NPV_SumExpression], bcd._before_npv)
            self.assertIs(bcd[LinearExpression], bcd._before_general_expression)
            # Named expression
            m.e = Expression(expr=m.p + m.q[0])
            self.assertEqual(bcd.register_dispatcher(visitor, m.e), (True, None))
            self.assertEqual(len(bcd), 7)
            self.assertIs(bcd[m.e.__class__], bcd._before_named_expression)

        finally:
            linear._before_child_dispatcher = _orig_dispatcher

    def test_to_expression(self):
        m = ConcreteModel()
        m.x = Var()
        m.y = Var()

        cfg = VisitorConfig()
        visitor = LinearRepnVisitor(**cfg)
        # prepopulate the visitor's var_map
        visitor.walk_expression(m.x + m.y)

        expr = LinearRepn()
        self.assertEqual(expr.to_expression(visitor), 0)

        expr.linear[id(m.x)] = 0
        self.assertEqual(expr.to_expression(visitor), 0)

        expr.linear[id(m.x)] = 1
        assertExpressionsEqual(self, expr.to_expression(visitor), m.x)

        expr.linear[id(m.x)] = 2
        assertExpressionsEqual(self, expr.to_expression(visitor), 2 * m.x)

        expr.linear[id(m.y)] = 3
        assertExpressionsEqual(self, expr.to_expression(visitor), 2 * m.x + 3 * m.y)

        expr.multiplier = 10
        assertExpressionsEqual(
            self, expr.to_expression(visitor), (2 * m.x + 3 * m.y) * 10
        )
        expr.multiplier = 1

        expr.constant = 0
        expr.linear[id(m.x)] = 0
        expr.linear[id(m.y)] = 0
        assertExpressionsEqual(self, expr.to_expression(visitor), 0)

    @unittest.skipUnless(numpy_available, "Test requires numpy")
    def test_nonnumeric(self):
        m = ConcreteModel()
        m.p = Param(mutable=True, initialize=numpy.array([3]), domain=Any)
        m.e = Expression()

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(m.p)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 3)
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.nonlinear, None)

        m.p = numpy.array([3, 4])

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(m.p)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        self.assertStructuredAlmostEqual(
            repn.constant, InvalidNumber(numpy.array([3, 4]))
        )
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.nonlinear, None)

    def test_zero_elimination(self):
        m = ConcreteModel()
        m.x = Var(range(4))

        e = 0 * m.x[0] + 0 * m.x[1] * m.x[2] + 0 * log(m.x[3])

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(e)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(
            cfg.var_map,
            {
                id(m.x[0]): m.x[0],
                id(m.x[1]): m.x[1],
                id(m.x[2]): m.x[2],
                id(m.x[3]): m.x[3],
            },
        )
        self.assertEqual(
            cfg.var_order, {id(m.x[0]): 0, id(m.x[1]): 1, id(m.x[2]): 2, id(m.x[3]): 3}
        )
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {})
        assertExpressionsEqual(self, repn.nonlinear, 0 * log(m.x[3]))

        m.p = Param(mutable=True, within=Any, initialize=None)
        e = m.p * m.x[0] + m.p * m.x[1] * m.x[2] + m.p * log(m.x[3])

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(e)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(
            cfg.var_map,
            {
                id(m.x[0]): m.x[0],
                id(m.x[1]): m.x[1],
                id(m.x[2]): m.x[2],
                id(m.x[3]): m.x[3],
            },
        )
        self.assertEqual(
            cfg.var_order, {id(m.x[0]): 0, id(m.x[1]): 1, id(m.x[2]): 2, id(m.x[3]): 3}
        )
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {id(m.x[0]): InvalidNumber(None)})
        self.assertEqual(repn.nonlinear, InvalidNumber(None))

        e = m.p * m.x[0] + m.p * m.x[1] + m.p * log(m.x[3])
        f = 1 + m.x[0] + 0 * e

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(f)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(
            cfg.var_map,
            {
                id(m.x[0]): m.x[0],
                id(m.x[1]): m.x[1],
                id(m.x[2]): m.x[2],
                id(m.x[3]): m.x[3],
            },
        )
        self.assertEqual(
            cfg.var_order, {id(m.x[0]): 0, id(m.x[1]): 1, id(m.x[2]): 2, id(m.x[3]): 3}
        )
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 1)
        self.assertEqual(len(repn.linear), 2)
        self.assertEqual(
            repn.linear,
            {id(m.x[0]): InvalidNumber(None), id(m.x[1]): InvalidNumber(None)},
        )
        self.assertEqual(repn.nonlinear, InvalidNumber(None))

        f = 1 + m.p + 0 * e

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(f)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(
            cfg.var_map,
            {
                id(m.x[0]): m.x[0],
                id(m.x[1]): m.x[1],
                id(m.x[2]): m.x[2],
                id(m.x[3]): m.x[3],
            },
        )
        self.assertEqual(
            cfg.var_order, {id(m.x[0]): 0, id(m.x[1]): 1, id(m.x[2]): 2, id(m.x[3]): 3}
        )
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, InvalidNumber(None))
        self.assertEqual(len(repn.linear), 2)
        self.assertEqual(
            repn.linear,
            {id(m.x[0]): InvalidNumber(None), id(m.x[1]): InvalidNumber(None)},
        )
        self.assertEqual(repn.nonlinear, InvalidNumber(None))

    def test_var_order(self):
        m = ConcreteModel()
        m.x = Var(range(4))

        e = 2 * m.x[1] + m.x[0] + 4 * m.x[3] + 3 * m.x[2]

        cfg = VisitorConfig()
        repn = LinearRepnVisitor(**cfg).walk_expression(e)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(
            cfg.var_map,
            {
                id(m.x[0]): m.x[0],
                id(m.x[1]): m.x[1],
                id(m.x[2]): m.x[2],
                id(m.x[3]): m.x[3],
            },
        )
        self.assertEqual(
            cfg.var_order, {id(m.x[0]): 0, id(m.x[1]): 1, id(m.x[2]): 2, id(m.x[3]): 3}
        )
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(len(repn.linear), 4)
        self.assertEqual(
            repn.linear, {id(m.x[0]): 1, id(m.x[1]): 2, id(m.x[2]): 3, id(m.x[3]): 4}
        )
        self.assertEqual(repn.nonlinear, None)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import enum


#
# pyomo - A pyomo.core.PyomoModel object, or a *.py file that defines such an object
# cpxlp - A CPLEX LP file
# nl - AMPL *.nl file
# mps - A free-format MPS file
# mod - AMPL *.mod file
# lpxlp - A LPSolve LP file
# osil - An XML file defined by the COIN-OR OS project: Instance
# bar - A Baron input file
# gams - A GAMS input file
#
class ProblemFormat(str, enum.Enum):
    pyomo = 'pyomo'
    cpxlp = 'cpxlp'
    nl = 'nl'
    mps = 'mps'
    mod = 'mod'
    lpxlp = 'lpxlp'
    osil = 'osil'
    bar = 'bar'
    gams = 'gams'

    # Overloading __str__ is needed to match the behavior of the old
    # pyutilib.enum class (removed June 2020). There are spots in the
    # code base that expect the string representation for items in the
    # enum to not include the class name. New uses of enum shouldn't
    # need to do this.
    def __str__(self):
        return self.value


#
# osrl - osrl XML file defined by the COIN-OR OS project: Result
# results - A Pyomo results object  (reader define by solver class)
# sol - AMPL *.sol file
# soln - A solver-specific solution file  (reader define by solver class)
# yaml - A Pyomo results file in YAML format
# json - A Pyomo results file in JSON format
#
class ResultsFormat(str, enum.Enum):
    osrl = 'osrl'
    results = 'results'
    sol = 'sol'
    soln = 'soln'
    yaml = 'yaml'
    json = 'json'

    # Overloading __str__ is needed to match the behavior of the old
    # pyutilib.enum class (removed June 2020). There are spots in the
    # code base that expect the string representation for items in the
    # enum to not include the class name. New uses of enum shouldn't
    # need to do this.
    def __str__(self):
        return self.value


def guess_format(filename):
    formats = {}
    formats['py'] = ProblemFormat.pyomo
    formats['nl'] = ProblemFormat.nl
    formats['bar'] = ProblemFormat.bar
    formats['mps'] = ProblemFormat.mps
    formats['mod'] = ProblemFormat.mod
    formats['lp'] = ProblemFormat.cpxlp
    formats['osil'] = ProblemFormat.osil
    formats['gms'] = ProblemFormat.gams
    formats['gams'] = ProblemFormat.gams

    formats['sol'] = ResultsFormat.sol
    formats['osrl'] = ResultsFormat.osrl
    formats['soln'] = ResultsFormat.soln
    formats['yml'] = ResultsFormat.yaml
    formats['yaml'] = ResultsFormat.yaml
    formats['jsn'] = ResultsFormat.json
    formats['json'] = ResultsFormat.json
    formats['results'] = ResultsFormat.yaml
    if filename:
        return formats.get(filename.split('.')[-1].strip(), None)
    else:
        return None
